# coding: utf-8

from importlib import import_module

from speedinfo.settings import speedinfo_settings


class ConditionsDispatcher:
    def __init__(self):
        self.conditions = None

    def import_conditions(self):
        self.conditions = []

        for module_path in speedinfo_settings.SPEEDINFO_PROFILING_CONDITIONS:
            path, class_name = module_path.rsplit('.', 1)

            try:
                module = import_module(path)
                self.conditions.append(
                    getattr(module, class_name)()
                )
            except (AttributeError, ImportError) as e:
                msg = 'Could not import "{}". {}: {}.'.format(module_path, e.__class__.__name__, e)
                raise ImportError(msg)

    def get_conditions(self):
        if self.conditions is None:
            self.import_conditions()

        return self.conditions


conditions_dispatcher = ConditionsDispatcher()



# coding: utf-8

from itertools import islice
from timeit import default_timer

from django.db import connection
from speedinfo import profiler
from speedinfo.conditions import conditions_dispatcher
from speedinfo.settings import speedinfo_settings

try:
    from django.urls import resolve, Resolver404  # Django >= 1.10
except ImportError:
    from django.core.urlresolvers import resolve, Resolver404


class ProfilerMiddleware(object):
    """
    Collects request and response statistics and saves profiler data.
    Unified middleware for all Django versions.
    """
    def __init__(self, get_response=None):
        self.get_response = get_response
        self.is_active = False
        self.force_debug_cursor = False
        self.start_time = 0
        self.existing_sql_count = 0

    def get_view_name(self, request):
        """Returns full view name from request, eg. 'app.module.view_name'.

        :param request: Request object
        :type request: :class:`django.http.HttpRequest`
        :return: view name or None if name can't be resolved
        :rtype: str or None
        """
        try:
            return resolve(request.path)._func_path
        except Resolver404:
            return None

    def can_process_request(self, request):
        """Checks conditions to start profiling the request

        :type request: :class:`django.http.HttpRequest`
        :return: True if request can be processed
        :rtype: bool
        """
        result = profiler.is_on and hasattr(request, 'user') and self.get_view_name(request)

        for condition in conditions_dispatcher.get_conditions():
            result = result and condition.process_request(request)

            if not result:
                break

        return result

    def can_process_response(self, response):
        """Checks conditions to finish profiling the request

        :type response: :class:`django.http.HttpResponse`
        :return: True if request can be processed
        :rtype: bool
        """
        result = True

        for condition in conditions_dispatcher.get_conditions():
            result = result and condition.process_response(response)

            if not result:
                break

        return result

    def process_view(self, request, *args, **kwargs):
        """Initialize statistics variables and environment.

        :return: Response object or None
        :rtype: :class:`django.http.HttpResponse` or None
        """
        self.is_active = self.can_process_request(request)

        if self.is_active:
            # Force DB connection to debug mode to get sql time and number of queries
            self.force_debug_cursor = connection.force_debug_cursor
            connection.force_debug_cursor = True

            self.existing_sql_count = len(connection.queries)
            self.start_time = default_timer()

    def process_response(self, request, response):
        """Aggregates request and response statistics and saves it in profiler data.

        :param request: Request object
        :type request: :class:`django.http.HttpRequest`
        :param response: Response object returned by a Django view or by a middleware
        :type response: :class:`django.http.HttpResponse` or :class:`django.http.StreamingHttpResponse`
        :return: View response
        :rtype: :class:`django.http.HttpResponse` or :class:`django.http.StreamingHttpResponse`
        """
        if self.is_active:
            if self.can_process_response(response):
                view_execution_time = default_timer() - self.start_time

                # Calculate the execution time and the number of queries.
                # Exclude queries made before the call of our middleware (e.g. in SessionMiddleware).
                sql_count = max(len(connection.queries) - self.existing_sql_count, 0)
                sql_time = sum(float(q['time']) for q in islice(connection.queries, self.existing_sql_count, None))

                # Collects request and response params
                view_name = self.get_view_name(request)
                is_anon_call = request.user.is_anonymous() if callable(request.user.is_anonymous) else request.user.is_anonymous
                is_cache_hit = getattr(response, speedinfo_settings.SPEEDINFO_CACHED_RESPONSE_ATTR_NAME, False)

                # Saves profiler data
                profiler.data.add(view_name, request.method, is_anon_call, is_cache_hit, sql_time, sql_count, view_execution_time)

            # Rollback debug cursor value even if process response condition is disabled
            connection.force_debug_cursor = self.force_debug_cursor

        return response

    def __call__(self, request):
        """New middleware handler introduced in Django 1.10.

        :param request: Request object
        :type request: :class:`django.http.HttpRequest`
        :return: View response
        :rtype: :class:`django.http.HttpResponse` or :class:`django.http.StreamingHttpResponse`
        """
        response = self.process_view(request)

        if response is None:
            response = self.get_response(request)

        return self.process_response(request, response)



# coding: utf-8

import re

from speedinfo.conditions.base import Condition
from speedinfo.settings import speedinfo_settings


class ExcludeURLCondition(Condition):
    """
    Plugin for conditional profiling based on the list of
    urls specified in SPEEDINFO_EXCLUDE_URLS settings
    """
    def __init__(self):
        self.patterns = None

    def get_patterns(self):
        if (self.patterns is None) or speedinfo_settings.SPEEDINFO_TESTS:
            self.patterns = [re.compile(pattern) for pattern in speedinfo_settings.SPEEDINFO_EXCLUDE_URLS]

        return self.patterns

    def process_request(self, request):
        """Checks requested page url against the list of excluded urls.

        :type request: :class:`django.http.HttpRequest`
        :return: False if path matches to any of the exclude urls
        :rtype: bool
        """
        for pattern in self.get_patterns():
            if pattern.match(request.path):
                return False

        return True

    def process_response(self, response):
        return True



# coding: utf-8

import django
from django.conf import settings
from django.core.checks import Error, Warning, register
from django.apps import AppConfig

if django.VERSION < (1, 10):
    MIDDLEWARE_SETTINGS_NAME = 'MIDDLEWARE_CLASSES'
else:
    MIDDLEWARE_SETTINGS_NAME = 'MIDDLEWARE'


def check_middleware(app_configs, **kwargs):
    profiler_middleware_index = None
    profiler_middleware_multiple = False
    cache_middleware_index = None
    errors = []

    for i, middleware in enumerate(getattr(settings, MIDDLEWARE_SETTINGS_NAME)):
        if middleware == 'speedinfo.middleware.ProfilerMiddleware':
            profiler_middleware_multiple = profiler_middleware_index is not None
            profiler_middleware_index = i
        elif middleware == 'django.middleware.cache.FetchFromCacheMiddleware':
            cache_middleware_index = i

    if profiler_middleware_index is None:
        errors.append(
            Warning(
                'speedinfo.middleware.ProfilerMiddleware is missing from %s' % MIDDLEWARE_SETTINGS_NAME,
                hint='Add speedinfo.middleware.ProfilerMiddleware to %s' % MIDDLEWARE_SETTINGS_NAME,
                id='speedinfo.W001',
            )
        )

    elif (cache_middleware_index is not None) and (profiler_middleware_index > cache_middleware_index):
        errors.append(
            Error(
                'speedinfo.middleware.ProfilerMiddleware occurs after '
                'django.middleware.cache.FetchFromCacheMiddleware in %s' % MIDDLEWARE_SETTINGS_NAME,
                hint='Move speedinfo.middleware.ProfilerMiddleware to before '
                     'django.middleware.cache.FetchFromCacheMiddleware in %s' % MIDDLEWARE_SETTINGS_NAME,
                id='speedinfo.E001',
            )
        )

    elif profiler_middleware_multiple:
        errors.append(
            Error(
                'speedinfo.middleware.ProfilerMiddleware occurs multiple times in %s' % MIDDLEWARE_SETTINGS_NAME,
                hint='Add speedinfo.middleware.ProfilerMiddleware only once in %s' % MIDDLEWARE_SETTINGS_NAME,
                id='speedinfo.E002',
            )
        )

    return errors


def check_cache_backend(app_configs, **kwargs):
    for params in settings.CACHES.values():
        if params.get('BACKEND') == 'speedinfo.backends.proxy_cache':
            return []

    return [
        Error(
            'speedinfo.backends.proxy_cache is missing from CACHES',
            hint='Add speedinfo.backends.proxy_cache to CACHES',
            id='speedinfo.E003',
        )
    ]


class SpeedinfoConfig(AppConfig):
    name = 'speedinfo'

    def ready(self):
        register()(check_middleware)
        register()(check_cache_backend)



# coding: utf-8

from django.db import models


class ViewProfiler(models.Model):
    """
    Holds profiler stats grouped by view.
    """
    view_name = models.CharField('View name', max_length=255)
    method = models.CharField('HTTP method', max_length=8)
    anon_calls = models.PositiveIntegerField('Anonymous calls', default=0)
    cache_hits = models.PositiveIntegerField('Cache hits', default=0)
    sql_total_time = models.FloatField('SQL total time', default=0)
    sql_total_count = models.PositiveIntegerField('SQL total queries count', default=0)
    total_calls = models.PositiveIntegerField('Total calls', default=0)
    total_time = models.FloatField('Total time', default=0)

    class Meta:
        verbose_name_plural = 'Views'
        unique_together = ('view_name', 'method')



# Generated by Django 2.0.4 on 2018-04-17 16:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('speedinfo', '0004_auto_20180409_1416'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='viewprofiler',
            options={'verbose_name_plural': 'Views'},
        ),
    ]



# coding: utf-8

import csv

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from django.conf.urls import url
from django.contrib import admin
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect, HttpResponse

from speedinfo import profiler
from speedinfo.models import ViewProfiler
from speedinfo.settings import speedinfo_settings

try:
    from django.urls import reverse  # Django >= 1.10
except ImportError:
    from django.core.urlresolvers import reverse


def field_wrapper(col):
    """Helper function to dynamically create list display method
    for :class:`ViewProfilerAdmin` to control value formating
    and sort order.

    :type col: :data:`settings.ReportColumnFormat`
    :rtype: function
    """
    def field_format(obj):
        return col.format.format(getattr(obj, col.attr_name))

    field_format.short_description = col.name
    field_format.admin_order_field = col.attr_name

    return field_format


class ViewProfilerAdmin(admin.ModelAdmin):
    list_display_links = None
    actions = None
    ordering = ('-total_time',)

    class Media:
        css = {
            'all': ('speedinfo/css/admin.css',)
        }

    def __init__(self, *args, **kwargs):
        """Initializes the list of visible columns and
        the way they are formating the values.
        """
        super(ViewProfilerAdmin, self).__init__(*args, **kwargs)
        self.list_display = []

        for rc in speedinfo_settings.SPEEDINFO_REPORT_COLUMNS_FORMAT:
            if rc.attr_name in speedinfo_settings.SPEEDINFO_REPORT_COLUMNS:
                method_name = '{}_wrapper'.format(rc.attr_name)
                setattr(self, method_name, field_wrapper(rc))
                self.list_display.append(method_name)

    def get_queryset(self, request):
        qs = super(ViewProfilerAdmin, self).get_queryset(request)

        for rc in speedinfo_settings.SPEEDINFO_REPORT_COLUMNS_FORMAT:
            if (rc.attr_name in speedinfo_settings.SPEEDINFO_REPORT_COLUMNS) and not isinstance(rc.expression, str):
                qs = qs.annotate(**{
                    rc.attr_name: rc.expression
                })

        return qs

    def change_view(self, *args, **kwargs):
        raise PermissionDenied

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def changelist_view(self, request, extra_context=None):
        return super(ViewProfilerAdmin, self).changelist_view(request, extra_context={
            'title': 'Views profiler',
            'profiler_is_on': profiler.is_on,
        })

    def get_urls(self):
        return [
            url(r'^switch/$', self.admin_site.admin_view(self.switch), name='speedinfo-profiler-switch'),
            url(r'^export/$', self.admin_site.admin_view(self.export), name='speedinfo-profiler-export'),
            url(r'^reset/$', self.admin_site.admin_view(self.reset), name='speedinfo-profiler-reset'),
        ] + super(ViewProfilerAdmin, self).get_urls()

    def switch(self, request):
        profiler.is_on = not profiler.is_on
        return HttpResponseRedirect(reverse('admin:speedinfo_viewprofiler_changelist'))

    def export(self, request):
        """Exports profiling data as a comma-separated file.

        :param request: :class:`django.http.HttpRequest`
        :return: comma-separated file
        :rtype: :class:`django.http.HttpResponse`
        """
        output = StringIO()
        export_columns = list(filter(lambda col: col.attr_name in speedinfo_settings.SPEEDINFO_REPORT_COLUMNS,
                                     speedinfo_settings.SPEEDINFO_REPORT_COLUMNS_FORMAT))

        csv_writer = csv.writer(output)
        csv_writer.writerow([col.name for col in export_columns])

        for row in self.get_queryset(request):
            csv_writer.writerow([
                col.format.format(getattr(row, col.attr_name))
                for col in export_columns
            ])

        response = HttpResponse(output.getvalue(), content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename=profiler.csv'

        return response

    def reset(self, request):
        profiler.data.reset()
        return HttpResponseRedirect(reverse('admin:speedinfo_viewprofiler_changelist'))


admin.site.register(ViewProfiler, ViewProfilerAdmin)



# coding: utf-8


class Condition:
    """
    Base class for user-defined profiling condition classes
    """
    def process_request(self, request):
        """
        :type request: :class:`django.http.HttpRequest`
        :return: False if requested page should be excluded from profiling
        :rtype: bool
        """
        raise NotImplementedError

    def process_response(self, response):
        """
        :type response: :class:`django.http.HttpResponse`
        :return: False if requested page should be excluded from profiling
        :rtype: bool
        """
        raise NotImplementedError



# coding: utf-8

from django.core.cache import InvalidCacheBackendError
from django.utils.module_loading import import_string

from speedinfo.settings import speedinfo_settings


def proxy_cache(location, params):
    backend = params.pop('CACHE_BACKEND')

    try:
        backend_cls = import_string(backend)
    except ImportError as e:
        raise InvalidCacheBackendError('Could not find backend "{}": {}'.format(backend, e))

    class ProxyCacheBackend(backend_cls):
        """
        Proxy access to cache backend to intercept requests to cached pages
        when using default Django per-site middleware or per-view cache decorator.
        """
        def get(self, key, default=None, version=None):
            response = super(ProxyCacheBackend, self).get(key, default, version)

            # Sets the flag for marking response from cache
            if (response is not None) and key.startswith('views.decorators.cache.cache_page'):
                setattr(response, speedinfo_settings.SPEEDINFO_CACHED_RESPONSE_ATTR_NAME, True)

            return response

    return ProxyCacheBackend(location, params)



# coding: utf-8

from os.path import join, dirname
from setuptools import setup


with open(join(dirname(__file__), 'README.rst')) as f:
    long_description = f.read()

setup(
    name='django-speedinfo',
    version='1.4.1',
    packages=['speedinfo', 'speedinfo.migrations', 'speedinfo.conditions'],
    include_package_data=True,
    license='MIT',
    description='Live profiling tool for Django framework to measure views performance',
    long_description=long_description,
    url='https://github.com/catcombo/django-speedinfo',
    author='Evgeniy Krysanov',
    author_email='evgeniy.krysanov@gmail.com',
    classifiers=[
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    keywords='django profiler views performance',
    install_requires=['Django>=1.8']
)



# coding: utf-8

from collections import namedtuple
from django.conf import settings
from django.db.models import ExpressionWrapper, F, FloatField, IntegerField

ReportColumnFormat = namedtuple('ReportColumnFormat', ['name', 'format', 'attr_name', 'expression'])

DEFAULTS = {
    'SPEEDINFO_TESTS': False,
    'SPEEDINFO_CACHED_RESPONSE_ATTR_NAME': 'is_cached',
    'SPEEDINFO_PROFILING_CONDITIONS': [
        'speedinfo.conditions.exclude_urls.ExcludeURLCondition',
    ],
    'SPEEDINFO_EXCLUDE_URLS': [],
    'SPEEDINFO_REPORT_COLUMNS': (
        'view_name', 'method', 'anon_calls_ratio', 'cache_hits_ratio',
        'sql_count_per_call', 'sql_time_ratio', 'total_calls', 'time_per_call', 'total_time'),
    'SPEEDINFO_REPORT_COLUMNS_FORMAT': [
        ReportColumnFormat('View name', '{}', 'view_name', 'view_name'),
        ReportColumnFormat('HTTP method', '{}', 'method', 'method'),
        ReportColumnFormat('Anonymous calls', '{:.1f}%', 'anon_calls_ratio',
                           ExpressionWrapper(100.0 * F('anon_calls') / F('total_calls'), output_field=FloatField())),
        ReportColumnFormat('Cache hits', '{:.1f}%', 'cache_hits_ratio',
                           ExpressionWrapper(100.0 * F('cache_hits') / F('total_calls'), output_field=FloatField())),
        ReportColumnFormat('SQL queries per call', '{}', 'sql_count_per_call',
                           ExpressionWrapper(F('sql_total_count') / F('total_calls'), output_field=IntegerField())),
        ReportColumnFormat('SQL time', '{:.1f}%', 'sql_time_ratio',
                           ExpressionWrapper(100.0 * F('sql_total_time') / F('total_time'), output_field=FloatField())),
        ReportColumnFormat('Total calls', '{}', 'total_calls', 'total_calls'),
        ReportColumnFormat('Time per call', '{:.8f}', 'time_per_call',
                           ExpressionWrapper(F('total_time') / F('total_calls'), output_field=FloatField())),
        ReportColumnFormat('Total time', '{:.4f}', 'total_time', 'total_time'),
    ]
}


class SpeedinfoSettings:
    def __init__(self, defaults=None):
        self.defaults = defaults or DEFAULTS

    def __getattr__(self, name):
        if name not in self.defaults:
            raise AttributeError('Invalid setting: "{}"'.format(name))

        return getattr(settings, name, self.defaults.get(name))


speedinfo_settings = SpeedinfoSettings()
